import asyncio
import threading
from concurrent.futures import ThreadPoolExecutor, Future, wait, FIRST_COMPLETED
from typing import Dict, List, Any, Set, Tuple
from queue import Queue

from event_systems.base.threaded_protocols import InstancedThreaded
from event_systems.base.handler import Handler

from event_systems.common_expressions import (
    NO_SUBSCRIPTION_FOUND,
    subscription_failure,
    subscription_success,
)


class ThreadedInternalEventSystem(InstancedThreaded):
    instances: List[str] = []

    def __init__(self) -> None:
        self._name = self._auto_name()
        ThreadedInternalEventSystem.instances.append(self._name)

        self._setup_initial_state()

    def __deinit__(self) -> None:
        ThreadedInternalEventSystem.instances.remove(self._name)

    def _setup_initial_state(self) -> None:
        self._is_running = False
        self._lock = threading.Lock()
        self._subscriptions: Dict[str, List[Handler]] = {}
        self._event_queue: Queue[Tuple[str, Dict[str, Any]]] = Queue()

        self._futures_not_done: Set[Future[Any]] = set()
        self._futures_done: Set[Future[Any]] = set()

    # NOTE: The way we calculate worker count suggests, that starting the event system should be done after subscriptions have beend registered.
    def start(self) -> None:
        assert not self._is_running, "Event system is already running."
        self._is_running = True

        worker_count = self._calculate_worker_count()
        self._executor = ThreadPoolExecutor(
            max_workers=worker_count,
            thread_name_prefix=f"{self._name}_execution",
        )

        n = f"{self._name}"
        t = threading.Thread(
            name=n,
            target=self._execution_loop,
            args=[worker_count],
        )
        t.start()

    def stop(self) -> None:
        self._is_running = False
        self._event_queue.join()
        self._executor.shutdown(wait=True)
        self._setup_initial_state()

    def subscribe(self, event_name: str, fn: Handler) -> Dict[str, Any]:
        with self._lock:
            try:
                if event_name not in self._subscriptions:
                    self._subscriptions[event_name] = []
                self._subscriptions[event_name].append(fn)

                return subscription_success(event_name)
            except Exception as e:
                return subscription_failure(event_name, e)

    def post(self, event_name: str, event_data: Dict[str, Any]) -> None:
        assert self._is_running, "Event system is not running."
        if event_name not in self._subscriptions:
            raise ValueError(NO_SUBSCRIPTION_FOUND.format(event=event_name))
        self._event_queue.put((event_name, event_data))

    def get_subscriptions(self) -> Dict[str, List[Handler]]:
        return self._subscriptions

    def is_running(self) -> bool:
        return self._is_running

    def _calculate_worker_count(self) -> int:
        total_handlers = sum(len(handlers) for handlers in self._subscriptions.values())
        avg_handlers_per_event = max(
            1, total_handlers // len(self._subscriptions) if self._subscriptions else 1
        )

        return max(1, len(self._subscriptions) * avg_handlers_per_event)

    def _run_handler(self, handler: Handler, event_data: Dict[str, Any]) -> None:
        if not callable(handler):
            raise TypeError("Handler is not callable.")
        if asyncio.iscoroutinefunction(handler):
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try:
                loop.run_until_complete(handler(event_data))
            finally:
                loop.close()
        else:
            # For synchronous functions, call them directly in the worker thread
            handler(event_data)

    def process_all_events(self) -> None:
        # Wait for all events in the queue to be processed
        self._event_queue.join()

        # Wait for any currently running handlers to complete
        self._wait_for_all_futures_to_complete()

    def _wait_for_all_futures_to_complete(self) -> None:
        while self._futures_not_done or self._futures_done:
            # Process futures that are not done
            if self._futures_not_done:
                done, self._futures_not_done = wait(
                    self._futures_not_done, return_when=FIRST_COMPLETED
                )
                for future in done:
                    try:
                        future.result()
                    except Exception as e:
                        # Log the exception or handle it appropriately
                        print(f"Exception in future: {e}")
                    finally:
                        self._futures_done.add(future)

            # Clean up done futures
            self._cleanup_completed_futures()

    def _execution_loop(self, max_concurrent: int) -> None:
        with self._executor as executor:
            while self._is_running:
                if self._event_queue.empty():
                    continue
                event_publication = self._event_queue.get()

                # Not done futures grows with every submission
                event_type, event_data = event_publication
                if event_type in self._subscriptions:
                    for handler in self._subscriptions[event_type]:
                        self._futures_not_done.add(
                            executor.submit(self._run_handler, handler, event_data)
                        )

                self._event_queue.task_done()
                # Once it is bigger than max concurrent, separate all done futures out of it
                # and clean up to release memory
                if len(self._futures_not_done) >= max_concurrent:
                    done, self._futures_not_done = wait(
                        self._futures_not_done,
                        return_when=FIRST_COMPLETED,
                    )
                    self._futures_done.update(done)
                    self._cleanup_completed_futures()

    def _cleanup_completed_futures(self) -> None:
        completed_futures_to_remove: Set[Future[Any]] = set()

        for future in self._futures_done:
            if future.done():
                future.result()
                completed_futures_to_remove.add(future)

        # Remove completed futures from self._futures_done
        self._futures_done -= completed_futures_to_remove

    def _auto_name(self) -> str:
        if len(ThreadedInternalEventSystem.instances) == 0:
            return "event_system_0"
        if reusable_name := self.find_reusable_name(
            ThreadedInternalEventSystem.instances
        ):
            return reusable_name
        else:
            return f"event_system_{len(self.instances)}"

    def find_reusable_name(self, instances: List[str]) -> None | str:
        if len(instances) == 1:
            index = int(instances[0].split("_")[-1])
            if index != 0:
                return "event_system_0"
        indices = [int(name.split("_")[-1]) for name in instances]
        for i in range(len(indices) - 1):
            if indices[i] + 1 != indices[i + 1]:
                missing_index = indices[i] + 1
                return f"event_system_{missing_index}"
        return None



import pytest
from event_systems.instanced.threaded_event_system import ThreadedInternalEventSystem
from tests.helpers.dummy_handlers import dummy_handler


def test_two_instances_with_different_threads_dont_interfere(capsys: pytest.CaptureFixture[str]) -> None: 
    # given
    es_1 = ThreadedInternalEventSystem()
    es_2 = ThreadedInternalEventSystem()
    
    # when
    es_1.subscribe("some_event", dummy_handler)
    es_2.subscribe("some_event", dummy_handler)

    # then 
    assert len(es_1.get_subscriptions()) == 1
    assert len(es_2.get_subscriptions()) == 1

    # when
    es_1.start()
    es_2.start()

    es_1.post("some_event", {"dummy_data": "data 1"})
    es_1.process_all_events()
    es_2.post("some_event", {"dummy_data": "data 2"})
    es_2.process_all_events()

    es_1.stop()
    es_2.stop()

    # then
    assert len(es_1.get_subscriptions()) == 0
    assert len(es_2.get_subscriptions()) == 0
    out, _ = capsys.readouterr()
    assert out == "data 1\ndata 2\n"



from typing import Any, Dict, List, Protocol
from event_systems.base.handler import Handler

# TODO: Unsubscribe
class InstancedThreaded(Protocol):
    def start(self) -> None: ...

    def stop(self) -> None: ...

    def subscribe(self, event_name: str, fn: Handler) -> Dict[str, Any]: ...

    def post(self, event_name: str, event_data: Dict[str, Any]) -> None: ...

    def get_subscriptions(self) -> Dict[str, List[Handler]]: ...

    def process_all_events(self) -> None: ...

    def is_running(self) -> bool: ...



import asyncio
from typing import Optional, Protocol, Any, List, Dict, runtime_checkable

from event_systems.base.handler import Handler

# TODO: Write Documentation
# TODO: Unsubscribe

@runtime_checkable
class InstancedAsync(Protocol):
    async def start(self) -> None: ...

    async def stop(self) -> None: ...

    async def subscribe(self, event_name: str, fn: Handler) -> Dict[str, Any]: ...

    async def post(self, event_name: str, event_data: Dict[str, Any]) -> None: ...

    async def get_subscriptions(self) -> Dict[str, List[Handler]]: ...

    async def process_all_events(self) -> None: ...

    async def is_running(self) -> bool: ...

    def get_loop(self) -> asyncio.AbstractEventLoop: ...


class SingletonAsync(Protocol):
    @classmethod
    async def start(cls) -> None: ...

    @classmethod
    async def stop(cls) -> None: ...

    @classmethod
    async def subscribe(cls, event_name: str, fn: Handler) -> Dict[str, Any]: ...

    @classmethod
    async def post(cls, event_name: str, event_data: Dict[str, Any]) -> None: ...

    @classmethod
    async def get_subscriptions(cls) -> Dict[str, List[Handler]]: ...

    @classmethod
    async def process_all_events(cls) -> None: ...

    @classmethod
    async def is_running(cls) -> bool: ...

    @classmethod
    async def get_instance(cls) -> Optional[object]: ...



import asyncio
from typing import Callable, Dict, List, Any, Tuple, cast

from event_systems.base.async_protocols import InstancedAsync
from event_systems.base.handler import Handler

from event_systems.common_expressions import (
    NO_SUBSCRIPTION_FOUND,
    subscription_failure,
    subscription_success,
)


class AsyncInternalEventSystem(InstancedAsync):
    """
    InternalEventSystem: An asyncio based, instanced event system.

    ATTENTION: Custom loop mechanism is experimental and not really working.
    This class provides an event system using asyncio for asynchronous event handling.
    It allows subscribing to events, posting events, and managing the event loop.
    """

    def __init__(self, asyncio_loop: asyncio.AbstractEventLoop | None = None) -> None:
        self._setup_initial_state(asyncio_loop)

    def _setup_initial_state(
        self,
        asyncio_loop: asyncio.AbstractEventLoop | None,
    ) -> None:
        self._is_running = False
        self._lock = asyncio.Lock()
        self._asyncio_loop = (
            asyncio_loop if asyncio_loop is not None else asyncio.get_event_loop()
        )
        self._subscriptions: Dict[str, List[Handler]] = {}
        self._event_queue: asyncio.Queue[Tuple[str, Dict[str, Any]]] = asyncio.Queue()

    async def start(self) -> None:
        assert self._is_running is False, "Event system is already running."
        assert hasattr(self, "_asyncio_loop"), "Event system has no asyncio loop."
        assert self._asyncio_loop, "Asyncio loop is None."

        self._is_running = True
        self._task = self._asyncio_loop.create_task(self._run_event_loop())

    async def stop(self) -> None:
        self._is_running = False
        if hasattr(self, "_event_queue"):
            await self._event_queue.join()

        if hasattr(self, "_task"):
            self._task.cancel()
            try:
                # Run wait_for in the context of the custom loop but don't await the result
                self._asyncio_loop.run_until_complete(
                    asyncio.wait_for(self._task, timeout=1.0)
                )
            except asyncio.CancelledError:
                pass  # Task was successfully cancelled
            except asyncio.TimeoutError:
                pass

        # Clean up resources
        if hasattr(self, "_event_queue"):
            del self._event_queue
        if hasattr(self, "_task"):
            del self._task

        # Reset state
        self._setup_initial_state(self._asyncio_loop)

    async def subscribe(self, event_name: str, fn: Handler) -> Dict[str, Any]:
        async with self._lock:
            try:
                if event_name not in self._subscriptions:
                    self._subscriptions[event_name] = []
                self._subscriptions[event_name].append(fn)

                return subscription_success(event_name)
            except Exception as e:
                return subscription_failure(event_name, e)

    async def post(self, event_name: str, event_data: Dict[str, Any]) -> None:
        if event_name not in self._subscriptions:
            raise ValueError(NO_SUBSCRIPTION_FOUND.format(event=event_name))

        await self._event_queue.put((event_name, event_data))

    async def get_subscriptions(self) -> Dict[str, List[Handler]]:
        return self._subscriptions

    async def is_running(self) -> bool:
        return self._is_running

    def get_loop(self) -> asyncio.AbstractEventLoop:
        return self._asyncio_loop

    async def process_all_events(self) -> None:
        if not hasattr(self, "_event_queue"):
            return
        # Ensure we're in the correct loop context
        if asyncio.get_running_loop() != self._asyncio_loop:
            raise RuntimeError(
                "This method must be called from the same loop as the event system was initialized with."
            )

        # Since we're already in the correct loop context, we can simply await
        await self._event_queue.join()

    async def _run_handler(self, handler: Handler, event_data: Dict[str, Any]) -> None:
        if asyncio.iscoroutinefunction(handler):
            # If it's an async function, await it
            await handler(event_data)
        else:
            # If it's a sync function, run it in a thread
            await asyncio.to_thread(cast(Callable[[Any], Any], handler), event_data)

    async def _process_events(self) -> None:
        while hasattr(self, "_event_queue"):
            event_type, event_data = await self._event_queue.get()
            if event_type in self._subscriptions:
                for handler in self._subscriptions[event_type]:
                    await self._run_handler(handler, event_data)
            self._event_queue.task_done()

    async def _run_event_loop(self) -> None:
        while self._is_running:
            await self._process_events()
            await asyncio.sleep(0.1)  # Prevent busy waiting



from typing import Any, Dict


NEEDS_INITIALIZATION = "'{class_name}' must be initialized before posting events."
NO_SUBSCRIPTION_FOUND = "No subscription found with '{event}'."


def subscription_success(event_name: str) -> Dict[str, Any]:
    return {
        "success": True,
        "message": f"Successfully subscribed to event: {event_name}",
    }


def subscription_failure(event_name: str, e: Exception) -> Dict[str, Any]:
    return {
        "success": False,
        "message": f"Failed to subscribe to event: {event_name}. Error: {str(e)}",
    }



import asyncio
from typing import Any, Coroutine
import pytest


from event_systems.instanced.async_event_system import AsyncInternalEventSystem
from tests.helpers.dummy_handlers import dummy_handler


def run_in_loop(
    coro: Coroutine[Any, Any, None],
    loop: asyncio.AbstractEventLoop,
) -> None:
    """Helper function to run a coroutine in a specified loop from the test context."""
    # Since run_until_complete is meant for synchronous context, we don't await it
    loop.run_until_complete(coro)


@pytest.mark.asyncio
async def test_two_internal_event_systems_are_independent() -> None:
    # given & when
    es_1 = AsyncInternalEventSystem()
    es_2 = AsyncInternalEventSystem()
    await es_1.subscribe("some_event", dummy_handler)

    # then different InterrnralEventSystem (instance) objects hold different subscriptions
    assert id(es_1) != id(es_2)
    assert len(await es_1.get_subscriptions()) != len(await es_2.get_subscriptions())


@pytest.mark.asyncio
async def test_custom_loop_stop_results_in_clean_state() -> None:
    # given
    custom_loop = asyncio.new_event_loop()
    es = AsyncInternalEventSystem(asyncio_loop=custom_loop)
    await es.subscribe("some_event", dummy_handler)
    await es.start()

    # when
    await es.stop()

    # then
    assert len(await es.get_subscriptions()) == 0
    assert await es.is_running() == False
    assert not hasattr(es, "_task")


@pytest.mark.asyncio
async def test_system_with_custom_loop_can_be_reused(
    capsys: pytest.CaptureFixture[str],
) -> None:
    # given
    custom_loop = asyncio.new_event_loop()
    try:
        es = AsyncInternalEventSystem(asyncio_loop=custom_loop)
        await es.subscribe("some_event", dummy_handler)
        await es.start()
        await es.stop()

        # when
        expected = "some different data"
        await es.subscribe("some_other_event", dummy_handler)
        await es.start()

        await es.subscribe("some_different_event", dummy_handler)
        await es.post("some_different_event", {"dummy_data": expected})
        run_in_loop(es.process_all_events(), custom_loop)

        # then
        out, _ = capsys.readouterr()
        assert out == expected + "\n"
        assert len(await es.get_subscriptions()) == 2
        assert await es.is_running() == True
        assert hasattr(es, "_task")
    finally:
        # Reset to the default loop for pytest
        asyncio.set_event_loop(asyncio.get_event_loop())
        # Ensure all tasks are cancelled before closing the loop
        for task in asyncio.all_tasks(custom_loop):
            task.cancel()
        custom_loop.run_until_complete(custom_loop.shutdown_asyncgens())
        custom_loop.close()


@pytest.mark.asyncio
async def test_get_loop_returns_correct_loop() -> None:
    # given
    custom_loop = asyncio.new_event_loop()
    try:
        es = AsyncInternalEventSystem(asyncio_loop=custom_loop)

        # when
        loop = es.get_loop()

        # then
        assert loop is custom_loop
    finally:
        asyncio.set_event_loop(asyncio.get_event_loop())
        # Ensure all tasks are cancelled before closing the loop
        for task in asyncio.all_tasks(custom_loop):
            task.cancel()
        custom_loop.run_until_complete(custom_loop.shutdown_asyncgens())
        custom_loop.close()



import asyncio
import contextlib
from typing import Callable, Optional, Dict, List, Any, Tuple, cast

from event_systems.base.async_protocols import SingletonAsync
from event_systems.base.handler import Handler

from event_systems.common_expressions import (
    NEEDS_INITIALIZATION,
    NO_SUBSCRIPTION_FOUND,
    subscription_success,
    subscription_failure,
)


class AsyncSharedEventSystem(SingletonAsync):
    _instance: Optional["AsyncSharedEventSystem"] = None
    _lock = asyncio.Lock()

    _is_running: bool
    _subscriptions: Dict[str, List[Handler]]
    _event_queue: asyncio.Queue[Tuple[str, Dict[str, Any]]]

    @classmethod
    async def start(cls) -> None:
        cls._is_running = True
        async with cls._lock:
            if not cls._instance:
                await cls._initialize()
        cls._task = asyncio.create_task(cls._run_event_loop())

    @classmethod
    async def stop(cls) -> None:
        # Wait for all items in the queue to be processed, then remove
        if hasattr(cls, "_event_queue"):
            await cls._event_queue.join()
            del cls._event_queue

        # Cancel the task if it exists, then remove
        if hasattr(cls, "_task") and cls._task:
            cls._task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await cls._task
            del cls._task

        # Reset state
        cls._instance = None
        cls._subscriptions = {}
        cls._is_running = False

    @classmethod
    async def subscribe(cls, event_name: str, fn: Handler) -> Dict[str, Any]:
        if not cls._instance:
            await cls._initialize()

        async with cls._lock:
            try:
                if event_name not in cls._subscriptions:
                    cls._subscriptions[event_name] = []
                cls._subscriptions[event_name].append(fn)
                return subscription_success(event_name)
            except Exception as e:
                return subscription_failure(event_name, e)

    @classmethod
    async def post(cls, event_name: str, event_data: Dict[str, Any]) -> None:
        if cls._instance is None:
            raise RuntimeError(NEEDS_INITIALIZATION.format(class_name=cls.__name__))
        if event_name not in cls._subscriptions:
            raise ValueError(NO_SUBSCRIPTION_FOUND.format(event=event_name))

        await cls._event_queue.put((event_name, event_data))

    @classmethod
    async def get_subscriptions(cls) -> Dict[str, List[Handler]]:
        return cls._subscriptions

    @classmethod
    async def is_running(cls) -> bool:
        return cls._is_running

    @classmethod
    async def process_all_events(cls) -> None:
        if not hasattr(cls, "_event_queue"):
            return
        await cls._event_queue.join()

    @classmethod
    async def _initialize(cls) -> None:
        if not cls._instance:
            cls._instance = cls()
            cls._subscriptions = {}
            cls._event_queue = asyncio.Queue()

    @classmethod
    async def get_instance(cls) -> Optional["AsyncSharedEventSystem"]:
        return cls._instance

    @classmethod
    async def _run_handler(cls, handler: Handler, event_data: Dict[str, Any]) -> None:
        if asyncio.iscoroutinefunction(handler):
            # If it's an async function, await it
            await handler(event_data)
        else:
            # If it's a sync function, run it in a thread
            await asyncio.to_thread(cast(Callable[[Any], Any], handler), event_data)

    @classmethod
    async def _process_events(cls) -> None:
        while hasattr(cls, "_event_queue"):
            event_type, event_data = await cls._event_queue.get()
            if event_type in cls._subscriptions:
                for handler in cls._subscriptions[event_type]:
                    # if handler: TODO: Check if this is ok
                    await cls._run_handler(handler, event_data)
            cls._event_queue.task_done()

    @classmethod
    async def _run_event_loop(cls) -> None:
        while cls._is_running:
            await cls._process_events()
            await asyncio.sleep(0.1)  # Prevent busy waiting



import pytest
from event_systems.singleton.async_event_system import AsyncSharedEventSystem

from tests.helpers.dummy_handlers import dummy_handler


@pytest.mark.asyncio
async def test_two_async_shared_event_systems_are_identical(
    async_shared_event_system: AsyncSharedEventSystem,
) -> None:
    # given & when
    es_1 = async_shared_event_system
    es_2 = async_shared_event_system
    await es_1.subscribe("some_event", dummy_handler)

    # then SharedEventSystem (singleton) references same object and hold same subscriptions
    assert id(es_1) == id(es_2)
    assert len(await es_1.get_subscriptions()) == len(await es_2.get_subscriptions())


@pytest.mark.asyncio
async def test_subscribe_when_not_initialized_calls_initialize(
    uninitialized_async_shared_event_system: AsyncSharedEventSystem,
) -> None:
    # given
    es = uninitialized_async_shared_event_system

    # when
    await es.subscribe("some_event", dummy_handler)

    # then
    assert await es.get_instance() is not None


@pytest.mark.asyncio
async def test_post_raises_exception_if_not_initialized(
    uninitialized_async_shared_event_system: AsyncSharedEventSystem,
) -> None:
    # given
    es = uninitialized_async_shared_event_system

    # when & then
    with pytest.raises(RuntimeError):
        await es.post("some_event", {"dummy_data": "some data"})
